use bincode::{self, options, Options};
use std::{net::{UdpSocket, SocketAddr}, io::{Read, Seek}};
use serde::{Serialize, Deserialize};
use sha2::{Digest,Sha512};
use regex::Regex;
use crc32fast;
use rand::{rngs::StdRng, RngCore, SeedableRng};
use num::{BigUint, one};

mod big_array;
mod primality_test;
mod prime_utils;
use big_array::BigArray;

use crate::prime_utils::prime_utils::is_prime_default;

const MAX_FRAME_PAYLOAD:u16=508;
const MAX_FRAME_PAYLOAD_U:usize=MAX_FRAME_PAYLOAD as usize;
const HEADER_SIZE:u16 = 12;
const MAX_PAYLOAD:u16 = MAX_FRAME_PAYLOAD - HEADER_SIZE;
const MAX_PAYLOAD_U:usize = MAX_PAYLOAD as usize;

#[derive(Serialize, Deserialize)]
struct PacketInformation {
    packet_numbers: u32, //4 bytes
    last_packet_size: u16, //2 bytes
    response_filename_checksum: u64, //8 bytes
} //14 bytes

#[derive(Serialize, Deserialize)]
struct Packet {
    packet_number: u32, //4 bytes
    payload_crc32: [u8; 4], //4 bytes
    payload_sum_crc32: u32, //4 bytes
    #[serde(with = "BigArray")]
    payload: [u8; MAX_PAYLOAD_U],
} //508 bytes

#[derive(Serialize, Deserialize)]
struct StrPacket {
    #[serde(with = "BigArray")]
    payload: [u8; MAX_PAYLOAD_U]
}

fn pow_mod(num: BigUint, pow: BigUint, modulo: BigUint) -> BigUint {
    let mut result = BigUint::from(1u8);
    let mut i = BigUint::from(0u8);
    while i < pow {
        result = (result * &num) % &modulo;
        i += one::<BigUint>();
    }

    result
}


//TODO: make this faster
fn new_p() -> BigUint {
    let mut private_key = [0u8; 128];
    let mut rng = StdRng::from_entropy();
    rng.fill_bytes(&mut private_key);
    let mut num = BigUint::from_bytes_be(&private_key);
    if is_prime_default(&num) {
        return num;
    } else {
        let higher: BigUint;
        loop {
            num += one::<BigUint>();
            if is_prime_default(&num) {
                higher = num;
                break;
            }
        }
        num = BigUint::from_bytes_be(&private_key);
        let lower: BigUint;
        loop {
            num -= one::<BigUint>();
            if is_prime_default(&num) {
                lower = num.clone();
                break;
            }
        }
        if &higher - &num > &num - &lower {
            return lower;
        } else {
            return higher;
        }
    }
}

fn main() {
    let port = "1337";
    let timeout = 100; //ms
    let local_addr: SocketAddr = ("0.0.0.0:".to_string()+port).parse().expect("Failed to parse address");
    let socket = UdpSocket::bind(local_addr).expect("Failed to bind socket");

    socket.set_read_timeout(Some(std::time::Duration::from_millis(timeout))).expect("set_read_timeout call failed");

    println!("UDP Server up and running on port {}", local_addr.port());

    let options = options().with_big_endian().allow_trailing_bytes().with_fixint_encoding();

    let hash_request_regex = Regex::new(r"[\w.-_ ]+:").unwrap();

    let missing_packet_request_regex = Regex::new(r"([\w.-_ ]+)(/\d{1,10})+").unwrap();

    loop {
        let mut buffer = [0u8; MAX_FRAME_PAYLOAD_U];
        let res = socket.recv_from(&mut buffer);
        if let Ok((_received_bytes, remote_addr)) = res {
            let filled_buffer = &buffer;//[..received_bytes];

            let request_packet = bincode::deserialize::<StrPacket>(filled_buffer).unwrap();
            let mut request: String = request_packet.payload.iter().map(|&c| c as char).collect();
            request.retain(|c| c != '\0'); //remove NUL bytes

            println!("Received request: {}", request);
            let req = request.as_str();

            if hash_request_regex.is_match(req) {
                println!("hash request");

                let mut sha512_hasher = Sha512::new();

                let file_result = std::fs::File::open("./files/".to_string() + req);

                if file_result.is_err() {
                    println!("File not found");
                    continue;
                }

                let mut file = file_result.unwrap();

                let mut file_buffer = [0u8; MAX_PAYLOAD_U];

                //read file chunk by chunk

                while let Ok(bytes_read) = file.read(&mut file_buffer) {
                    if bytes_read == 0 {
                        break;
                    }
                    sha512_hasher.update(&file_buffer[..bytes_read]);
                }

                let result = sha512_hasher.finalize();
                //add 1 byte of padding to result
                let mut result_bytes = [0u8; 65];
                result_bytes[..64].copy_from_slice(&result[..64]);

                socket.send_to(&result_bytes, remote_addr).expect("Failed to send file hash");

                continue;
            }

            if missing_packet_request_regex.is_match(req) {
                println!("missing packet request");
                let mut split = req.split("/");
                let filename = split.next().unwrap();

                let possible_file = std::fs::File::open("./files/".to_string() + filename);
                if possible_file.is_err() {
                    //println!("File not found");
                    continue;
                }
                let mut file = possible_file.unwrap();

                for packet_number in split {
                    file.seek(std::io::SeekFrom::Start((packet_number.parse::<u64>().unwrap() * MAX_PAYLOAD as u64) as u64)).unwrap();
                    let mut file_buffer = [0u8; MAX_PAYLOAD_U];
                    let bytes_read = file.read(&mut file_buffer).unwrap();
                    let mut packet = Packet {
                        packet_number: packet_number.parse::<u32>().unwrap(),
                        payload_crc32: crc32fast::hash(&file_buffer[..bytes_read]).to_be_bytes(),
                        payload_sum_crc32: 0,
                        payload: [0u8; MAX_PAYLOAD_U],
                    };
                    packet.payload[..bytes_read].copy_from_slice(&file_buffer[..bytes_read]);
                    let packet_bytes = options.serialize(&packet).unwrap();
                    socket.send_to(&packet_bytes, remote_addr).expect("Failed to send packet");
                }

                continue;
            }

            println!("Received file request");

            let file_result = std::fs::File::open("./files/".to_string() + req);

            if file_result.is_err() {
                println!("File not found");
                continue;
            }

            let mut file = file_result.unwrap();

            let mut file_buffer = [0u8; MAX_PAYLOAD_U];

            let mut packet_number = 0u32;

            let file_length = file.metadata().unwrap().len();
            println!("file length: {}", file_length);

            let packet_numbers = f64::ceil(file_length as f64/MAX_PAYLOAD as f64) as u32;

            let packet_information = PacketInformation {
                packet_numbers,
                last_packet_size: (file_length%MAX_PAYLOAD as u64) as u16,
                response_filename_checksum: request_packet.payload.iter().fold(0u64, |acc, &x| acc.wrapping_add(x as u64)),
            };

            socket.send_to(options.serialize(&packet_information).unwrap().as_slice(), remote_addr).expect("Failed to send packet information");

            let mut sha512_hasher = Sha512::new();

            //read file chunk by chunk

            let mut hasher = crc32fast::Hasher::new();

            while let Ok(bytes_read) = file.read(&mut file_buffer) {
                if bytes_read == 0 {
                    break;
                }

                hasher.update(&file_buffer[..bytes_read]);

                let mut packet = Packet {
                    packet_number,
                    payload_crc32: crc32fast::hash(&file_buffer[..bytes_read]).to_be_bytes(),
                    payload: [0u8; MAX_PAYLOAD_U],
                    payload_sum_crc32: hasher.clone().finalize(),
                };
                packet.payload[..bytes_read].copy_from_slice(&file_buffer[..bytes_read]);
                let packet_bytes = options.serialize(&packet).unwrap();
                socket.send_to(&packet_bytes, remote_addr).expect("Failed to send packet");
                packet_number += 1;

                sha512_hasher.update(&file_buffer[..bytes_read]);
            }

            let result = sha512_hasher.finalize();
            //add 1 byte of padding to result
            let mut result_bytes = [0u8; 65];
            result_bytes[..64].copy_from_slice(&result[..64]);

            socket.send_to(&result_bytes, remote_addr).expect("Failed to send file hash");
        }
    }
}
